class Provider(object):
    def connect(self):
        raise NotImplementedError('Method connect(self) is not implemented.')

    def close(self):
        raise NotImplementedError('Method close(self) is not implemented.')

    def get_tables(self):
        raise NotImplementedError('Method get_tables(self) is not implemented.')

    def clean_table(self, table_name):
        raise NotImplementedError('Method clean_table(self, table_name) is not implemented.')

    def clear_table_column(self, table_name, column_name):
        raise NotImplementedError('Method clear_table_column(self, table_name, column_name) is not implemented.')

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class TransactionScope(object):
    def __init__(self, database):
        self._database = database
        self._cursor = cursor

    def begin(self):
        self._cursor = self.database.cursor()

    def rollback(self):
        self._cursor.rollback()

    def commit(self):
        self._cursor.commit()

    def execute(self, query):
        return self._cursor.execute(query)

    def __enter__(self):
        self.begin()

    def __exit__(self):
        try:
            self.commit()
        except:
            self.rollback()


import yaml

from .definition import Definition
from .table_definition import TableDefinition

class DefinitionFileReader(object):
    def read(self, filepath):
        definition = Definition()

        with open(filepath, 'r') as stream:
            try:
                data = yaml.load(stream)

                for table in data.get('tables', {}).iteritems():
                    definition.tables.append(TableDefinition.from_yaml(table))
            except yaml.YAMLError as error:
                print(error)

        return definition


class ExecutionStep(object):
    def run(self, db):
        raise NotImplementedError('Method run(self, db) is not implemented.')

    def explain(self):
        raise NotImplementedError('Method explain(self) is not implemented.')


class Definition(object):
    def __init__(self):
        self.tables = []

    def __repr__(self):
        return '<Definition>\n{tables}\n</Definition>'.format(
            tables='\n'.join([str(table) for table in self.tables]),
        )


class TableFieldDefinition(object):
    def __init__(self, name, action):
        self.name = name
        self.action = action

    @staticmethod
    def from_yaml(data):
        field, options = data

        return TableFieldDefinition(
            name=field,
            action=options.get('action', None),
        )

    def __repr__(self):
        return '<Field name={name} action={action} />'.format(
            name=self.name,
            action=self.action,
        )


from .execution_step import ExecutionStep

class ExecutionPlan(object):
    def __init__(self, name=None):
        self.name = name
        self.execution_steps = []

    def add_step(self, step):
        if not isinstance(step, ExecutionPlan) and not isinstance(step, ExecutionStep):
            raise Exception('Execution step must be an instance of either ExecutionPlan or ExecutionStep.')

        self.execution_steps.append(step)

    def run(self, db):
        for step in self.execution_steps:
            print('Running: ' + step.explain())
            step.run(db)

        db._connection.commit()

    def explain(self):
        logs = []

        if self.name:
            logs.append(self.name + ':')

        for step in self.execution_steps:
            logs.append(step.explain())

        return '\n'.join(logs) + '\n'

    def __repr__(self):
        return '<ExecutionPlan>\n{steps}\n</ExecutionPlan>'.format(
            steps=self.execution_steps,
        )


class Column(object):
    def __init__(self, name, type, default_value, is_primary):
        self.name = name
        self.type = type
        self.default_value = default_value
        self.is_primary = is_primary

    def __repr__(self):
        return '<Column name={name} type={type} />'.format(
            name=self.name,
            type=self.type,
        )


from .execution_step import ExecutionStep

class ClearTableColumnExecutionStep(ExecutionStep):
    def __init__(self, table_name, column_name):
        self.table_name = table_name
        self.column_name = column_name

    def run(self, db):
        return db.clear_table_column(self.table_name, self.column_name)

    def explain(self):
        return 'Clear all columns values of {column} on table {table}.'.format(
            table=self.table_name,
            column=self.column_name,
        )